@@ -27,13 +27,11 @@
 #include "bamf-window.h"
 #include "bamf-legacy-screen.h"
 
-#ifdef HAVE_WEBAPPS
-#include "bamf-unity-webapps-application.h"
-#include "bamf-unity-webapps-tab.h"
-#endif
 #include <strings.h>
 
 #define BAMF_INDEX_NAME "bamf-2.index"
+#define EXEC_DESKTOP_FILE_OVERRIDE "--desktop_file_hint"
+#define ENV_DESKTOP_FILE_OVERRIDE "BAMF_DESKTOP_FILE_HINT"
 
 G_DEFINE_TYPE (BamfMatcher, bamf_matcher, BAMF_DBUS_TYPE_MATCHER_SKELETON);
 #define BAMF_MATCHER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE(obj, \
@@ -59,8 +57,9 @@ static guint matcher_signals[LAST_SIGNAL] = { 0 };
 const gchar* EXEC_BAD_PREFIXES[] =
 {
   "^gksu(do)?$", "^sudo$", "^su-to-root$", "^amdxdg-su$", "^java(ws)?$", "^cli$",
-  "^mono$", "^ruby$", "^padsp$", "^aoss$", "^python(\\d.\\d)?$", "^(ba)?sh$",
+  "^mono$", "^ruby$", "^padsp$", "^aoss$", "^python(\\d(\\.\\d)?)?$", "^(ba)?sh$",
   "^perl$", "^env$", "^xdg-open$", "^qmlscene$", "^qmlviewer$",
+  "^unity-webapps-runner$", "^webapp-container$",
   /* javaws strings: */ "^net\\.sourceforge\\.jnlp\\.runtime\\.Boot$", "^rt\\.jar$",
                         "^com\\.sun\\.javaws\\.Main$", "^deploy\\.jar$"
 };
@@ -71,7 +70,7 @@ const gchar* EXEC_BAD_SUFIXES = "(\\.bin|\\.py|\\.pl|\\.qml)$";
 // Prefixes that must be considered starting point of exec strings
 const gchar* EXEC_GOOD_PREFIXES[] =
 {
-  "^gnome-control-center$", "^libreoffice$", "^ooffice$", "^wine$", "^steam$",
+  "^unity-control-center$", "^libreoffice$", "^ooffice$", "^wine$", "^steam$",
   "^sol$"
 };
 
@@ -82,8 +81,6 @@ const gchar * CLASS_BAD_VALUES[] =
   "com-sun-javaws-Main", "VCLSalFrame"
 };
 
-const gchar * EXEC_DESKTOP_FILE_OVERRIDE = "--desktop_file_hint";
-
 static void
 on_view_active_changed (BamfView *view, gboolean active, BamfMatcher *matcher)
 {
@@ -151,7 +148,7 @@ bamf_matcher_get_application_by_desktop_file (BamfMatcher *self, const char *des
 
   g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
 
-  if (!desktop_file)
+  if (!desktop_file || desktop_file[0] == '\0')
     return NULL;
 
   for (l = self->priv->views; l; l = l->next)
@@ -559,6 +556,49 @@ get_exec_overridden_desktop_file (const char *exec_string)
   return result;
 }
 
+char *
+get_env_overridden_desktop_file (guint pid)
+{
+  gchar *environ_file;
+  gchar *environ;
+  gchar *result;
+  gsize file_len;
+  gint i;
+
+  if (pid < 2)
+    return NULL;
+
+  environ = NULL;
+  result = NULL;
+  environ_file = g_strdup_printf ("/proc/%u/environ", pid);
+
+  if (g_file_get_contents (environ_file, &environ, &file_len, NULL))
+    {
+      for (i = 0; i < file_len && environ && environ[i] != '\0'; i += strlen (environ + i) + 1)
+        {
+          const gchar *var = environ + i;
+
+          if (g_str_has_prefix (var, ENV_DESKTOP_FILE_OVERRIDE"="))
+            {
+              const gchar *file_path;
+              file_path = var + sizeof (ENV_DESKTOP_FILE_OVERRIDE);
+
+              if (g_str_has_suffix (file_path, ".desktop") &&
+                  g_file_test (file_path, G_FILE_TEST_EXISTS|G_FILE_TEST_IS_REGULAR))
+                {
+                  result = g_strdup (file_path);
+                  break;
+                }
+            }
+        }
+    }
+
+  g_free (environ);
+  g_free (environ_file);
+
+  return result;
+}
+
 static GList *
 pid_parent_tree (BamfMatcher *self, guint pid)
 {
@@ -1491,7 +1531,7 @@ is_web_app_window (BamfLegacyWindow *window)
   const char *window_class = bamf_legacy_window_get_class_name (window);
   const char *instance_name = bamf_legacy_window_get_class_instance_name (window);
 
-  // Chrome/Chromium uses url wm_class strings to represent its web apps.
+  // Chrome/Chromium uses url wm_class strings to represent its web apps (--app mode).
   // These apps will still have the same parent pid and hints as the main chrome
   // window, so we skip the hint check.
   // We can tell a window is a chrome web app window if its instance name is
@@ -1620,6 +1660,7 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
   BamfMatcherPrivate *priv;
   BamfLegacyWindow *window;
   GList *desktop_files = NULL, *l;
+  char *app_id;
   char *desktop_file = NULL;
   const char *desktop_class = NULL;
   const char *class_name = NULL;
@@ -1632,7 +1673,7 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
 
   priv = self->priv;
   window = bamf_window_get_window (bamf_window);
-  desktop_file = bamf_legacy_window_get_hint (window, _NET_WM_DESKTOP_FILE);
+  desktop_file = bamf_legacy_window_get_hint (window, _BAMF_DESKTOP_FILE);
   class_name = bamf_legacy_window_get_class_name (window);
   instance_name = bamf_legacy_window_get_class_instance_name (window);
 
@@ -1672,8 +1713,17 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
           g_free (desktop_file);
         }
     }
-  else
+
+  if (!desktop_files)
     {
+      gint pid = bamf_legacy_window_get_pid (window);
+      desktop_file = get_env_overridden_desktop_file (pid);
+
+      if (desktop_file)
+        {
+          desktop_files = g_list_prepend (desktop_files, desktop_file);
+        }
+
       const char *exec_string = bamf_legacy_window_get_exec_string (window);
       desktop_file = get_exec_overridden_desktop_file (exec_string);
 
@@ -1681,6 +1731,27 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
         {
           desktop_files = g_list_prepend (desktop_files, desktop_file);
         }
+
+      if (!desktop_files)
+        {
+          app_id = bamf_legacy_window_get_hint (window, _GTK_APPLICATION_ID);
+
+          if (app_id)
+            {
+              for (l = g_hash_table_lookup (priv->desktop_id_table, app_id); l; l = l->next)
+                {
+                  desktop_file = l->data;
+                  desktop_class = bamf_matcher_get_desktop_file_class (self, desktop_file);
+
+                  if ((!filter_by_wmclass && !desktop_class) || g_strcmp0 (desktop_class, target_class) == 0)
+                    {
+                      desktop_files = g_list_prepend (desktop_files, g_strdup (desktop_file));
+                    }
+                }
+
+              g_free (app_id);
+            }
+        }
     }
 
   desktop_file = NULL;
@@ -2004,7 +2075,7 @@ ensure_window_hint_set (BamfMatcher *self,
       return;
     }
 
-  desktop_file_hint = bamf_legacy_window_get_hint (window, _NET_WM_DESKTOP_FILE);
+  desktop_file_hint = bamf_legacy_window_get_hint (window, _BAMF_DESKTOP_FILE);
 
   if (desktop_file_hint)
     {
@@ -2040,7 +2111,7 @@ ensure_window_hint_set (BamfMatcher *self,
   g_list_free (pids);
 
   if (desktop_file_hint)
-    bamf_legacy_window_set_hint (window, _NET_WM_DESKTOP_FILE, desktop_file_hint);
+    bamf_legacy_window_set_hint (window, _BAMF_DESKTOP_FILE, desktop_file_hint);
 }
 
 static void
@@ -2133,7 +2204,7 @@ handle_raw_window (BamfMatcher *self, BamfLegacyWindow *window)
 }
 
 static char *
-get_gnome_control_center_window_hint (BamfMatcher * self, BamfLegacyWindow * window)
+get_unity_control_center_window_hint (BamfMatcher * self, BamfLegacyWindow * window)
 {
   const gchar *role;
   GList *list;
@@ -2145,21 +2216,21 @@ get_gnome_control_center_window_hint (BamfMatcher * self, BamfLegacyWindow * win
 
   if (role)
     {
-      gchar *exec = g_strconcat ("gnome-control-center ", role, NULL);
+      gchar *exec = g_strconcat ("unity-control-center ", role, NULL);
       list = g_hash_table_lookup (self->priv->desktop_file_table, exec);
       g_free (exec);
     }
 
   if (!role || !list)
     {
-      list = g_hash_table_lookup (self->priv->desktop_id_table, "gnome-control-center");
+      list = g_hash_table_lookup (self->priv->desktop_id_table, "unity-control-center");
     }
 
   return (list ? (char *) list->data : NULL);
 }
 
 static void
-on_gnome_control_center_window_role_changed (BamfLegacyWindow *window, BamfMatcher* self)
+on_unity_control_center_window_role_changed (BamfLegacyWindow *window, BamfMatcher* self)
 {
   g_return_if_fail (BAMF_IS_MATCHER (self));
   g_return_if_fail (BAMF_IS_LEGACY_WINDOW (window));
@@ -2167,8 +2238,8 @@ on_gnome_control_center_window_role_changed (BamfLegacyWindow *window, BamfMatch
   char *old_hint;
   const char *new_hint;
 
-  old_hint = bamf_legacy_window_get_hint (window, _NET_WM_DESKTOP_FILE);
-  new_hint = get_gnome_control_center_window_hint (self, window);
+  old_hint = bamf_legacy_window_get_hint (window, _BAMF_DESKTOP_FILE);
+  new_hint = get_unity_control_center_window_hint (self, window);
 
   if (new_hint && g_strcmp0 (new_hint, old_hint) != 0)
     {
@@ -2178,6 +2249,18 @@ on_gnome_control_center_window_role_changed (BamfLegacyWindow *window, BamfMatch
   g_free (old_hint);
 }
 
+static void
+handle_window_opening (BamfLegacyScreen *screen, const gchar *desktop_id, SnStartupSequence *startup_sequence, BamfMatcher *self)
+{
+  bamf_matcher_set_starting_desktop_file (self, desktop_id, startup_sequence, TRUE);
+}
+
+static void
+handle_window_opening_finished (BamfLegacyScreen *screen, const gchar *desktop_id, BamfMatcher *self)
+{
+  bamf_matcher_set_starting_desktop_file (self, desktop_id, NULL, FALSE);
+}
+
 static void
 handle_window_opened (BamfLegacyScreen * screen, BamfLegacyWindow * window, BamfMatcher *self)
 {
@@ -2200,17 +2283,17 @@ handle_window_opened (BamfLegacyScreen * screen, BamfLegacyWindow * window, Bamf
           return;
         }
     }
-  else if (g_strcmp0 (bamf_legacy_window_get_class_name (window), "Gnome-control-center") == 0)
+  else if (g_strcmp0 (bamf_legacy_window_get_class_name (window), "Unity-control-center") == 0)
     {
-      char *old_hint = bamf_legacy_window_get_hint (window, _NET_WM_DESKTOP_FILE);
-      const char *new_hint = get_gnome_control_center_window_hint (self, window);
+      char *old_hint = bamf_legacy_window_get_hint (window, _BAMF_DESKTOP_FILE);
+      const char *new_hint = get_unity_control_center_window_hint (self, window);
 
       if (new_hint && g_strcmp0 (old_hint, new_hint) != 0)
         {
-          bamf_legacy_window_set_hint (window, _NET_WM_DESKTOP_FILE, new_hint);
+          bamf_legacy_window_set_hint (window, _BAMF_DESKTOP_FILE, new_hint);
         }
 
-      g_signal_connect (window, "role-changed", (GCallback) on_gnome_control_center_window_role_changed, self);
+      g_signal_connect (window, "role-changed", (GCallback) on_unity_control_center_window_role_changed, self);
       g_free (old_hint);
     }
 
@@ -2332,6 +2415,34 @@ is_autostart_desktop_file (const gchar *desktop_file)
   return autostart;
 }
 
+gboolean
+is_valid_desktop_file (const gchar *desktop_file)
+{
+  if (!desktop_file || desktop_file[0] == '\0')
+    return FALSE;
+
+  if (!g_file_test (desktop_file, G_FILE_TEST_EXISTS))
+    return FALSE;
+
+  return TRUE;
+}
+
+gboolean
+is_visible_desktop_file (BamfMatcher *self,
+                         const gchar *desktop_file)
+{
+  if (!is_valid_desktop_file (desktop_file))
+    return FALSE;
+
+  if (is_no_display_desktop (self, desktop_file))
+    return FALSE;
+
+  if (is_autostart_desktop_file (desktop_file))
+    return FALSE;
+
+  return TRUE;
+}
+
 void
 bamf_matcher_register_desktop_file_for_pid (BamfMatcher * self,
                                             const gchar * desktop_file,
@@ -2339,28 +2450,81 @@ bamf_matcher_register_desktop_file_for_pid (BamfMatcher * self,
 {
   gpointer key;
   BamfLegacyScreen *screen;
-  GList *windows, *l;
+  GList *windows, *pids, *l, *ll;
+  gboolean matching_window_found;
 
   g_return_if_fail (BAMF_IS_MATCHER (self));
   g_return_if_fail (desktop_file);
 
-  if (is_autostart_desktop_file (desktop_file))
+  if (!is_visible_desktop_file (self, desktop_file))
     return;
 
   key = GUINT_TO_POINTER (pid);
   g_hash_table_insert (self->priv->registered_pids, key, g_strdup (desktop_file));
 
-  /* fixme, this is a bit heavy */
-
   screen = bamf_legacy_screen_get_default ();
-
   g_return_if_fail (BAMF_IS_LEGACY_SCREEN (screen));
 
   windows = bamf_legacy_screen_get_windows (screen);
+  matching_window_found = FALSE;
 
   for (l = windows; l; l = l->next)
     {
-      ensure_window_hint_set (self, l->data);
+      BamfLegacyWindow *window = l->data;
+
+      if (!BAMF_IS_LEGACY_WINDOW (window))
+        continue;
+
+      pids = pid_parent_tree (self, bamf_legacy_window_get_pid (window));
+
+      for (ll = pids; ll; ll = ll->next)
+        {
+          if (pid == GPOINTER_TO_UINT (ll->data))
+            {
+              matching_window_found = TRUE;
+              break;
+            }
+        }
+
+      g_list_free (pids);
+
+      if (matching_window_found)
+        {
+          ensure_window_hint_set (self, window);
+          break;
+        }
+    }
+}
+
+void
+bamf_matcher_set_starting_desktop_file (BamfMatcher *self,
+                                        const char *desktop_file,
+                                        SnStartupSequence *startup_sequence,
+                                        gboolean starting)
+{
+  BamfApplication *app;
+
+  g_return_if_fail (BAMF_IS_MATCHER (self));
+  g_return_if_fail (desktop_file != NULL);
+
+  if (!is_visible_desktop_file (self, desktop_file))
+    return;
+
+  app = bamf_matcher_get_application_by_desktop_file (self, desktop_file);
+
+  if (!BAMF_IS_APPLICATION (app) && starting)
+    {
+      app = bamf_application_new_from_desktop_file (desktop_file);
+
+      if (!bamf_matcher_is_view_registered (self, BAMF_VIEW (app)))
+        {
+          bamf_matcher_register_view_stealing_ref (self, BAMF_VIEW (app));
+        }
+    }
+
+  if (BAMF_IS_APPLICATION (app))
+    {
+      bamf_view_set_starting (BAMF_VIEW (app), startup_sequence, starting);
     }
 }
 
@@ -2939,103 +3103,6 @@ on_dbus_handle_window_stack_for_monitor (BamfDBusMatcher *interface,
   return TRUE;
 }
 
-#ifdef HAVE_WEBAPPS
-static gboolean
-bamf_matcher_has_tab_with_parent_xid (BamfMatcher *matcher, guint64 xid)
-{
-  GList *l;
-  g_return_val_if_fail (BAMF_IS_MATCHER (matcher), FALSE);
-
-  for (l = matcher->priv->views; l; l = l->next)
-    {
-      if (!BAMF_IS_TAB (l->data))
-        continue;
-
-      if (xid == bamf_tab_get_xid (BAMF_TAB (l->data)))
-        return TRUE;
-    }
-
-  return FALSE;
-}
-
-static void
-on_webapp_child_added (BamfView *application,
-                       BamfView *child,
-                       gpointer user_data)
-{
-  BamfMatcher *self;
-  BamfLegacyWindow *legacy_window;
-  BamfUnityWebappsTab *webapp_tab;
-
-  g_return_if_fail (BAMF_IS_MATCHER (user_data));
-  g_return_if_fail (BAMF_IS_UNITY_WEBAPPS_TAB (child));
-
-  self = BAMF_MATCHER (user_data);
-  webapp_tab = BAMF_UNITY_WEBAPPS_TAB (child);
-  legacy_window = bamf_unity_webapps_tab_get_legacy_window_for (webapp_tab);
-
-  if (legacy_window && is_web_app_window (legacy_window))
-    {
-      /* If we have a chromeless window, we remove the window from the
-       * application children list, so that it won't be duplicated in launcher */
-
-      guint tab_xid = bamf_tab_get_xid (BAMF_TAB (webapp_tab));
-
-      if (!bamf_matcher_has_tab_with_parent_xid (self, tab_xid))
-        {
-          BamfApplication *old_application = bamf_matcher_get_application_by_xid (self, tab_xid);
-
-          if (BAMF_IS_APPLICATION (old_application))
-            {
-              BamfWindow *bamf_window = bamf_application_get_window (old_application, tab_xid);
-
-              if (BAMF_IS_VIEW (bamf_window))
-                bamf_view_remove_child (BAMF_VIEW (old_application), BAMF_VIEW (bamf_window));
-            }
-        }
-    }
-
-  bamf_matcher_register_view_stealing_ref (self, child);
-}
-
-static void on_webapp_child_removed (BamfView *application,
-                                     BamfView *child,
-                                     gpointer user_data)
-{
-  BamfLegacyWindow *legacy_window;
-  BamfUnityWebappsTab *webapp_tab;
-
-  g_return_if_fail (BAMF_IS_UNITY_WEBAPPS_TAB (child));
-
-  webapp_tab = BAMF_UNITY_WEBAPPS_TAB (child);
-  legacy_window = bamf_unity_webapps_tab_get_legacy_window_for (webapp_tab);
-
-  if (is_web_app_window (legacy_window))
-    {
-      /* If we have a chromeless window, we re-match it again as soon as the
-       * webapp handler is gone, so that we don't lose its control */
-      bamf_legacy_window_reopen (legacy_window);
-    }
-}
-
-static void
-on_webapp_appeared (BamfUnityWebappsObserver *observer,
-                    BamfApplication *application,
-                    gpointer user_data)
-{
-  BamfMatcher *self;
-
-  self = (BamfMatcher *)user_data;
-
-  bamf_matcher_register_view_stealing_ref (self, (BamfView *)application);
-
-  g_signal_connect (application, "child-added-internal", G_CALLBACK (on_webapp_child_added), self);
-  g_signal_connect (application, "child-removed-internal", G_CALLBACK (on_webapp_child_removed), self);
-
-  bamf_unity_webapps_application_add_existing_interests (BAMF_UNITY_WEBAPPS_APPLICATION (application));
-}
-#endif
-
 static void
 bamf_matcher_init (BamfMatcher * self)
 {
@@ -3069,11 +3136,20 @@ bamf_matcher_init (BamfMatcher * self)
                              &(priv->desktop_class_table));
 
   screen = bamf_legacy_screen_get_default ();
+  g_signal_connect (G_OBJECT (screen), BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_OPENING,
+                    G_CALLBACK (handle_window_opening), self);
+
+  g_signal_connect (G_OBJECT (screen), BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_OPENING_COMPLETED,
+                    G_CALLBACK (handle_window_opening_finished), self);
+
+  g_signal_connect (G_OBJECT (screen), BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_OPENING_CANCELED,
+                    G_CALLBACK (handle_window_opening_finished), self);
+
   g_signal_connect (G_OBJECT (screen), BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_OPENED,
-                    (GCallback) handle_window_opened, self);
+                    G_CALLBACK (handle_window_opened), self);
 
   g_signal_connect (G_OBJECT (screen), BAMF_LEGACY_SCREEN_SIGNAL_STACKING_CHANGED,
-                    (GCallback) handle_stacking_changed, self);
+                    G_CALLBACK (handle_stacking_changed), self);
 
   XSetErrorHandler (x_error_handler);
 
@@ -3116,13 +3192,6 @@ bamf_matcher_init (BamfMatcher * self)
 
   g_signal_connect (self, "handle-window-stack-for-monitor",
                     G_CALLBACK (on_dbus_handle_window_stack_for_monitor), self);
-
-#ifdef HAVE_WEBAPPS
-  priv->webapps_observer = bamf_unity_webapps_observer_new ();
-
-  g_signal_connect (priv->webapps_observer, "application-appeared",
-                    G_CALLBACK (on_webapp_appeared), self);
-#endif
 }
 
 static void
@@ -3136,14 +3205,6 @@ bamf_matcher_dispose (GObject *object)
       bamf_matcher_unregister_view (self, priv->views->data);
     }
 
-#ifdef HAVE_WEBAPPS
-  if (priv->webapps_observer)
-    {
-      g_object_unref (G_OBJECT (priv->webapps_observer));
-      priv->webapps_observer = NULL;
-    }
-#endif
-
   G_OBJECT_CLASS (bamf_matcher_parent_class)->dispose (object);
 }
 
