@@ -45,7 +45,8 @@
 #define BAMF_IS_MATCHER_CLASS(klass)            (G_TYPE_CHECK_CLASS_TYPE ((klass), BAMF_TYPE_MATCHER))
 #define BAMF_MATCHER_GET_CLASS(obj)             (G_TYPE_INSTANCE_GET_CLASS ((obj), BAMF_TYPE_MATCHER, BamfMatcherClass))
 
-#define _NET_WM_DESKTOP_FILE "_NET_WM_DESKTOP_FILE"
+#define _BAMF_DESKTOP_FILE "_BAMF_DESKTOP_FILE"
+#define _GTK_APPLICATION_ID "_GTK_APPLICATION_ID"
 
 typedef struct _BamfMatcher BamfMatcher;
 typedef struct _BamfMatcherClass BamfMatcherClass;
@@ -73,6 +74,11 @@ void          bamf_matcher_register_desktop_file_for_pid (BamfMatcher * self,
                                                           const char *application,
                                                           guint64 pid);
 
+void          bamf_matcher_set_starting_desktop_file     (BamfMatcher * self,
+                                                          const char *desktop_file,
+                                                          SnStartupSequence *sequence,
+                                                          gboolean starting);
+
 const char  * bamf_matcher_get_desktop_file_class        (BamfMatcher * self,
                                                           const char * desktop_file);
 
